@@ -44,6 +44,7 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <efivar.h>
+#include <inttypes.h>
 #include "list.h"
 #include "efi.h"
 #include "efichar.h"
@@ -72,12 +73,6 @@ static	LIST_HEAD(boot_entry_list);
 static	LIST_HEAD(blk_list);
 efibootmgr_opt_t opts;
 
-static inline void
-var_num_from_name(const char *pattern, char *name, uint16_t *num)
-{
-	sscanf(name, pattern, num);
-}
-
 static void
 free_vars(list_t *head)
 {
@@ -115,6 +110,13 @@ read_vars(char **namelist,
 					       &entry->attributes);
 			if (rc < 0)
 				goto err;
+
+			/* latest apple firmware sets high bit which appears
+			 * invalid to the linux kernel if we write it back so
+			 * lets zero it out if it is set since it would be
+			 * invalid to set it anyway */
+			entry->attributes = entry->attributes & ~(1 << 31);
+
 			entry->name = namelist[i];
 			list_add_tail(&entry->list, head);
 		}
@@ -143,12 +145,12 @@ static int
 compare(const void *a, const void *b)
 {
 	int rc = -1;
-	uint32_t n1, n2;
+	uint16_t n1, n2;
 	memcpy(&n1, a, sizeof(n1));
 	memcpy(&n2, b, sizeof(n2));
 	if (n1 < n2) rc = -1;
 	if (n1 == n2) rc = 0;
-	if (n2 > n2) rc = 1;
+	if (n1 > n2) rc = 1;
 	return rc;
 }
 
@@ -282,7 +284,7 @@ static int
 read_boot_order(efi_variable_t **boot_order)
 {
 	int rc;
-	efi_variable_t *new = NULL, *bo;
+	efi_variable_t *new = NULL, *bo = NULL;
 
 	if (*boot_order == NULL) {
 		new = calloc(1, sizeof (**boot_order));
@@ -298,6 +300,14 @@ read_boot_order(efi_variable_t **boot_order)
 	if (rc < 0 && new != NULL) {
 		free(new);
 		*boot_order = NULL;
+		bo = NULL;
+	}
+
+	if (bo) {
+		/* latest apple firmware sets high bit which appears invalid
+		 * to the linux kernel if we write it back so lets zero it out
+		 * if it is set since it would be invalid to set it anyway */
+		bo->attributes = bo->attributes & ~(1 << 31);
 	}
 	return rc;
 }
@@ -320,6 +330,8 @@ add_to_boot_order(uint16_t num)
 	old_data = (uint16_t *)boot_order->data;
 	new_data_size = boot_order->data_size + sizeof(uint16_t);
 	new_data = malloc(new_data_size);
+	if (!new_data)
+		return -1;
 
 	new_data[0] = num;
 	memcpy(new_data+1, old_data, boot_order->data_size);
@@ -336,13 +348,64 @@ add_to_boot_order(uint16_t num)
 	return rc;
 }
 
+static int
+remove_dupes_from_boot_order(void)
+{
+	efi_variable_t *boot_order = NULL;
+	uint64_t new_data_size;
+	uint16_t *new_data, *old_data;
+	unsigned int old_i,new_i;
+	int rc;
+
+	rc = read_boot_order(&boot_order);
+	if (rc < 0)
+		return rc;
+
+	old_data = (uint16_t *)(boot_order->data);
+	/* Start with the same size */
+	new_data_size = boot_order->data_size;
+	new_data = malloc(new_data_size);
+	if (!new_data)
+		return -1;
+
+	unsigned int old_max = boot_order->data_size / sizeof(*new_data);
+	for (old_i = 0, new_i = 0; old_i < old_max; old_i++) {
+		int copies = 0;
+		unsigned int j;
+		for (j = 0; j < new_i; j++) {
+			if (new_data[j] == old_data[old_i]) {
+				copies++;
+				break;
+			}
+		}
+		if (copies == 0) {
+			/* Copy this value */
+			new_data[new_i] = old_data[old_i];
+			new_i++;
+		}
+	}
+	/* Adjust the size if we didn't copy everything. */
+	new_data_size = sizeof(new_data[0]) * new_i;
+
+	/* Now new_data has what we need */
+	free(boot_order->data);
+	boot_order->data = (uint8_t *)new_data;
+	boot_order->data_size = new_data_size;
+	efi_del_variable(EFI_GLOBAL_GUID, "BootOrder");
+	rc = efi_set_variable(EFI_GLOBAL_GUID, "BootOrder", boot_order->data,
+				boot_order->data_size, boot_order->attributes);
+	free(boot_order->data);
+	free(boot_order);
+	return rc;
+}
+
 static int
 remove_from_boot_order(uint16_t num)
 {
 	efi_variable_t *boot_order = NULL;
 	uint64_t new_data_size;
 	uint16_t *new_data, *old_data;
-	int old_i,new_i;
+	unsigned int old_i,new_i;
 	int rc;
 
 	rc = read_boot_order(&boot_order);
@@ -375,8 +438,10 @@ remove_from_boot_order(uint16_t num)
 	boot_order->data = (uint8_t *)new_data;
 	boot_order->data_size = new_data_size;
 	efi_del_variable(EFI_GLOBAL_GUID, "BootOrder");
-	return efi_set_variable(EFI_GLOBAL_GUID, "BootOrder", boot_order->data,
+	rc = efi_set_variable(EFI_GLOBAL_GUID, "BootOrder", boot_order->data,
 				boot_order->data_size, boot_order->attributes);
+	free(boot_order->data);
+	return rc;
 }
 
 static int
@@ -393,18 +458,21 @@ read_boot_u16(const char *name)
 	if (rc < 0)
 		return rc;
 	if (data_size != 2) {
+		if (data != NULL)
+			free(data);
 		errno = EINVAL;
 		return -1;
 	}
 
 	rc = data[0];
+	free(data);
 	return rc;
 }
 
 static int
 set_boot_u16(const char *name, uint16_t num)
 {
-	return efi_set_variable(EFI_GLOBAL_GUID, name, (uint8_t *)&num, 
+	return efi_set_variable(EFI_GLOBAL_GUID, name, (uint8_t *)&num,
 				sizeof (num), EFI_VARIABLE_NON_VOLATILE |
 					      EFI_VARIABLE_BOOTSERVICE_ACCESS |
 					      EFI_VARIABLE_RUNTIME_ACCESS);
@@ -446,7 +514,7 @@ delete_boot_var(uint16_t num)
 
 
 static void
-set_var_nums(const char *pattern, list_t *list)
+set_var_nums(list_t *list)
 {
 	list_t *pos;
 	efi_variable_t *var;
@@ -456,7 +524,7 @@ set_var_nums(const char *pattern, list_t *list)
 
 	list_for_each(pos, list) {
 		var = list_entry(pos, efi_variable_t, list);
-		rc = sscanf(var->name, pattern, &num);
+		rc = sscanf(var->name, "Boot%04X-%*s", &num);
 		if (rc == 1) {
 			var->num = num;
 			name = var->name; /* shorter name */
@@ -560,8 +628,11 @@ construct_boot_order(char *bootorder, int keep,
 	size_t data_size = 0;
 
 	rc = parse_boot_order(bootorder, (uint16_t **)&data, &data_size);
-	if (rc < 0 || data_size <= 0)
+	if (rc < 0 || data_size == 0) {
+		if (data) /* this can't actually happen, but covscan believes */
+			free(data);
 		return rc;
+	}
 
 	if (!keep) {
 		*ret_data = data;
@@ -577,10 +648,18 @@ construct_boot_order(char *bootorder, int keep,
 		return 0;
 	}
 
+	/* latest apple firmware sets high bit which appears invalid
+	 * to the linux kernel if we write it back so lets zero it out
+	 * if it is set since it would be invalid to set it anyway */
+	bo.attributes = bo.attributes & ~(1 << 31);
+
 	size_t new_data_size = data_size + bo.data_size;
 	uint16_t *new_data = calloc(1, new_data_size);
-	if (!new_data)
+	if (!new_data) {
+		if (data)
+			free(data);
 		return -1;
+	}
 
 	memcpy(new_data, data, data_size);
 	memcpy(new_data + (data_size / sizeof (*new_data)), bo.data,
@@ -620,7 +699,7 @@ set_boot_order(int keep_old_entries)
 
 	rc = construct_boot_order(opts.bootorder, keep_old_entries,
 				(uint16_t **)&data, &data_size);
-	if (rc < 0 || data_size < 0)
+	if (rc < 0 || data_size == 0)
 		return rc;
 
 	return efi_set_variable(EFI_GLOBAL_GUID, "BootOrder", data, data_size,
@@ -693,12 +772,8 @@ show_boot_vars()
 				printf("\n");
 				continue;
 			}
-			if (optional_data_len > path->length + 4) {
-				printf("(invalid optional data length)\n");
-				continue;
-			}
 
-			rc = unparse_raw_text(text_path, text_path_len,
+			rc = unparse_raw_text(NULL, 0,
 				((uint8_t *)path)
 					+ load_option->file_path_list_length,
 				optional_data_len);
@@ -738,17 +813,27 @@ show_boot_order()
 	rc = read_boot_order(&boot_order);
 
 	if (rc < 0) {
-		perror("show_boot_order()");
-		return;
+		if (errno == ENOENT) {
+			boot_order = calloc(1, sizeof (*boot_order));
+			if (!boot_order) {
+				perror("show_boot_order()");
+				return;
+			}
+		} else {
+			perror("show_boot_order()");
+			return;
+		}
 	}
 
 	/* We've now got an array (in boot_order->data) of the
 	   boot order.  First add our entry, then copy the old array.
 	*/
 	data = (uint16_t *)boot_order->data;
-	if (boot_order->data_size)
+	if (boot_order->data_size) {
 		unparse_boot_order(data, boot_order->data_size / sizeof(uint16_t));
-
+		free(boot_order->data);
+	}
+	free(boot_order);
 }
 
 static int
@@ -811,6 +896,7 @@ usage()
 	printf("\t-b | --bootnum XXXX   modify BootXXXX (hex)\n");
 	printf("\t-B | --delete-bootnum delete bootnum (hex)\n");
 	printf("\t-c | --create         create new variable bootnum and add to bootorder\n");
+	printf("\t-D | --remove-dups	remove duplicate values from BootOrder\n");
 	printf("\t-d | --disk disk       (defaults to /dev/sda) containing loader\n");
 	printf("\t-e | --edd [1|3|-1]   force EDD 1.0 or 3.0 creation variables, or guess\n");
 	printf("\t-E | --device num      EDD 1.0 device number (defaults to 0x80)\n");
@@ -857,7 +943,8 @@ set_default_opts()
 static void
 parse_opts(int argc, char **argv)
 {
-	int c, num, rc;
+	int c, rc;
+	unsigned int num;
 	int option_index = 0;
 
 	while (1)
@@ -870,6 +957,7 @@ parse_opts(int argc, char **argv)
 			{"bootnum",          required_argument, 0, 'b'},
 			{"delete-bootnum",         no_argument, 0, 'B'},
 			{"create",                 no_argument, 0, 'c'},
+			{"remove-dups",            no_argument, 0, 'D'},
 			{"disk",             required_argument, 0, 'd'},
 			{"iface",            required_argument, 0, 'i'},
 			{"acpi_hid",         required_argument, 0, 'H'},
@@ -899,7 +987,7 @@ parse_opts(int argc, char **argv)
 		};
 
 		c = getopt_long (argc, argv,
-				 "AaBb:cd:e:E:gH:i:l:L:n:No:Op:qt:TuU:v::Vw"
+				 "AaBb:cDd:e:E:gH:i:l:L:n:No:Op:qt:TuU:v::Vw"
 				 "@:h",
 				 long_options, &option_index);
 		if (c == -1)
@@ -921,7 +1009,7 @@ parse_opts(int argc, char **argv)
 			break;
 		case 'b':
 			rc = sscanf(optarg, "%X", &num);
-			if (rc == 1 && num >= 0 && num < 0xffff) {
+			if (rc == 1 && num < 0xffff) {
 				opts.bootnum = num;
 			} else {
 				fprintf (stderr,"invalid hex value %s\n",optarg);
@@ -931,11 +1019,14 @@ parse_opts(int argc, char **argv)
 		case 'c':
 			opts.create = 1;
 			break;
+		case 'D':
+			opts.deduplicate = 1;
+			break;
 		case 'd':
 			opts.disk = optarg;
 			break;
 		case 'e':
-			rc = sscanf(optarg, "%d", &num);
+			rc = sscanf(optarg, "%u", &num);
 			if (rc == 1) opts.edd_version = num;
 			else {
 				fprintf (stderr,"invalid numeric value %s\n",optarg);
@@ -1038,7 +1129,7 @@ parse_opts(int argc, char **argv)
 			if (optarg) {
 				if (!strcmp(optarg, "v"))  opts.verbose = 2;
 				if (!strcmp(optarg, "vv")) opts.verbose = 3;
-				rc = sscanf(optarg, "%d", &num);
+				rc = sscanf(optarg, "%u", &num);
 				if (rc == 1)  opts.verbose = num;
 				else {
 					fprintf (stderr,"invalid numeric value %s\n",optarg);
@@ -1083,7 +1174,10 @@ main(int argc, char **argv)
 		return 0;
 	}
 
-	if (opts.iface && opts.acpi_hid == -1 && opts.acpi_uid == -1) {
+	if (opts.iface && (
+			opts.acpi_hid < 0 || opts.acpi_uid < 0 ||
+			opts.acpi_hid > UINT32_MAX ||
+			opts.acpi_uid > UINT32_MAX)) {
 		fprintf(stderr, "\nYou must specify the ACPI HID and UID when using -i.\n\n");
 		return 1;
 	}
@@ -1095,7 +1189,7 @@ main(int argc, char **argv)
 
 	read_boot_var_names(&boot_names);
 	read_vars(boot_names, &boot_entry_list);
-	set_var_nums("Boot%04X-%*s", &boot_entry_list);
+	set_var_nums(&boot_entry_list);
 
 	if (opts.delete_boot) {
 		if (opts.bootnum == -1) {
@@ -1136,6 +1230,9 @@ main(int argc, char **argv)
 		ret = set_boot_order(opts.keep_old_entries);
 	}
 
+	if (opts.deduplicate) {
+		ret = remove_dupes_from_boot_order();
+	}
 
 	if (opts.delete_bootnext) {
 		ret = efi_del_variable(EFI_GLOBAL_GUID, "BootNext");
