@@ -65,7 +65,7 @@ enum _bus_type {bus_type_unknown, isa, pci};
 enum _interface_type {interface_type_unknown,
 		      ata, atapi, scsi, usb,
 		      i1394, fibre, i2o, md,
-		      virtblk};
+		      virtblk, nvme};
 
 
 unsigned int lcm(unsigned int x, unsigned int y);
