@@ -112,7 +112,7 @@ read_var_names(filter_t filter, char ***namelist)
 
 		newlist = tmp;
 	}
-	if (rc == 0) {
+	if (rc == 0 && newlist) {
 		qsort(newlist, nentries, sizeof (char *), cmpstringp);
 		*namelist = newlist;
 	} else {
@@ -309,7 +309,7 @@ make_pci_device_path(uint8_t bus, uint8_t device, uint8_t function,
 	struct pci_access *pacc;
 	struct list_head *pos, *n;
 	LIST_HEAD(pci_parent_list);
-	size_t needed;
+	ssize_t needed;
 	off_t buf_offset = 0;
 
 	pacc = pci_alloc();
@@ -351,6 +351,20 @@ make_pci_device_path(uint8_t bus, uint8_t device, uint8_t function,
 	return buf_offset;
 }
 
+static ssize_t
+make_nvme_device_path(uint32_t ns_id, uint8_t *buf, size_t size)
+{
+	NVME_DEVICE_PATH p;
+	memset(&p, 0, sizeof(p));
+	p.type = 3;
+	p.subtype = 23;
+	p.length = sizeof(p);
+	p.namespace_id = ns_id;
+	if (size >= p.length)
+		memcpy(buf, &p, p.length);
+	return p.length;
+}
+
 static ssize_t
 make_scsi_device_path(uint16_t id, uint16_t lun, uint8_t *buf, size_t size)
 {
@@ -420,13 +434,18 @@ make_edd30_device_path(int fd, uint8_t *buf, size_t size)
 	int rc=0, interface_type;
 	unsigned char bus=0, device=0, function=0;
 	Scsi_Idlun idlun;
+	uint32_t ns_id;
 	unsigned char host=0, channel=0, id=0, lun=0;
-	size_t needed;
+	ssize_t needed;
 	off_t buf_offset = 0;
 
 	rc = disk_get_pci(fd, &interface_type, &bus, &device, &function);
 	if (rc) return 0;
-	if (interface_type != virtblk) {
+	if (interface_type == nvme) {
+		rc = get_nvme_ns_id(fd, &ns_id);
+		if (rc)
+			return 0;
+	} else if (interface_type != virtblk) {
 		memset(&idlun, 0, sizeof(idlun));
 		rc = get_scsi_idlun(fd, &idlun);
 		if (rc) return 0;
@@ -444,7 +463,13 @@ make_edd30_device_path(int fd, uint8_t *buf, size_t size)
 		return needed;
 	buf_offset += needed;
 
-	if (interface_type != virtblk) {
+	if (interface_type == nvme) {
+		needed = make_nvme_device_path(ns_id, buf + buf_offset,
+					size == 0 ? 0 : size - buf_offset);
+		if (needed < 0)
+			return needed;
+		buf_offset += needed;
+	} else if (interface_type != virtblk) {
 		needed = make_scsi_device_path(id, lun, buf + buf_offset,
 					size == 0 ? 0 : size - buf_offset);
 		if (needed < 0)
@@ -483,7 +508,7 @@ make_disk_load_option(char *disk, uint8_t *buf, size_t size)
 	uint8_t mbr_type=0, signature_type=0;
 	uint64_t part_start=0, part_size=0;
 	efi_char16_t *os_loader_path;
-	size_t needed = 0;
+	ssize_t needed = 0;
 	off_t buf_offset = 0;
 
 	memset(signature, 0, sizeof(signature));
@@ -503,8 +528,10 @@ make_disk_load_option(char *disk, uint8_t *buf, size_t size)
 			needed = make_edd10_device_path(opts.edd10_devicenum,
 							buf, size);
 		}
-		if (needed < 0)
+		if (needed < 0) {
+			close(disk_fd);
 			return needed;
+		}
 		buf_offset += needed;
 	}
 
@@ -557,7 +584,7 @@ get_virt_pci(char *name, unsigned char *bus,
 	ssize_t lnksz;
 
 	if (snprintf(inbuf, sizeof inbuf, "/sys/bus/virtio/devices/%s",
-			name) >= sizeof inbuf) {
+			name) >= (ssize_t)(sizeof inbuf)) {
 		return -1;
 	}
 
@@ -586,11 +613,11 @@ static ssize_t
 make_net_load_option(char *iface, uint8_t *buf, size_t size)
 {
 	/* copied pretty much verbatim from the ethtool source */
-	int fd = 0, err; 
+	int fd = 0, err;
 	unsigned char bus, slot, func;
 	struct ifreq ifr;
 	struct ethtool_drvinfo drvinfo;
-	size_t needed;
+	ssize_t needed;
 	off_t buf_offset;
 
 	memset(&ifr, 0, sizeof(ifr));
@@ -606,13 +633,16 @@ make_net_load_option(char *iface, uint8_t *buf, size_t size)
 	err = ioctl(fd, SIOCETHTOOL, &ifr);
 	if (err < 0) {
 		perror("Cannot get driver information");
+		close(fd);
 		return -1;
 	}
 
 	if (strncmp(drvinfo.bus_info, "virtio", 6) == 0) {
 		err = get_virt_pci(drvinfo.bus_info, &bus, &slot, &func);
-		if (err < 0)
+		if (err < 0) {
+			close(fd);
 			return err;
+		}
 	} else {
 		/* The domain part was added in 2.6 kernels.
 		 * Test for that first. */
@@ -623,6 +653,7 @@ make_net_load_option(char *iface, uint8_t *buf, size_t size)
 						&bus, &slot, &func);
 			if (err != 3) {
 				perror("Couldn't parse device location string.");
+				close(fd);
 				return -1;
 			}
 		}
@@ -630,9 +661,11 @@ make_net_load_option(char *iface, uint8_t *buf, size_t size)
 
 	err = ioctl(fd, SIOCGIFHWADDR, &ifr);
 	if (err < 0) {
+		close(fd);
 		perror("Cannot get hardware address.");
 		return -1;
 	}
+	close(fd);
 
 	buf_offset = 0;
 	needed = make_acpi_device_path(opts.acpi_hid, opts.acpi_uid, buf,
@@ -656,7 +689,7 @@ make_net_load_option(char *iface, uint8_t *buf, size_t size)
 		return needed;
 	buf_offset += needed;
 
-	needed = make_end_device_path(buf + buf_offset, 
+	needed = make_end_device_path(buf + buf_offset,
 					size == 0 ? 0 : size - buf_offset);
 	if (needed < 0)
 		return needed;
@@ -689,10 +722,10 @@ make_linux_load_option(uint8_t **data, size_t *data_size)
 	size_t load_option_size = sizeof (*load_option);
 	efi_char16_t description[64];
 	uint8_t *buf;
-	size_t needed;
+	ssize_t needed;
 	off_t buf_offset = 0, desc_offset;
 
-	load_option = calloc(1, sizeof (load_option));
+	load_option = calloc(1, sizeof (*load_option));
 	if (load_option == NULL)
 		return -1;
 	buf = (uint8_t *)load_option;
@@ -721,11 +754,19 @@ make_linux_load_option(uint8_t **data, size_t *data_size)
 
 	if (opts.iface) {
 		needed = make_net_load_option(opts.iface, NULL, 0);
+		if (needed < 0) {
+			free(buf);
+			return needed;
+		}
 		buf = extend(load_option, load_option_size, needed);
 		make_net_load_option(opts.iface, buf + buf_offset, needed);
 		buf_offset += needed;
 	} else {
 		needed = make_disk_load_option(opts.iface, NULL, 0);
+		if (needed < 0) {
+			free(buf);
+			return needed;
+		}
 		buf = extend(load_option, load_option_size, needed);
 		make_disk_load_option(opts.iface, buf + buf_offset, needed);
 		buf_offset += needed;
@@ -751,7 +792,7 @@ append_extra_args_ascii(uint8_t **data, size_t *data_size)
 	int i;
 	unsigned long usedchars=0;
 
-	if (!data)
+	if (!data || *data)
 		return -1;
 
 	for (i=opts.optind; i < opts.argc; i++)	{
@@ -775,8 +816,6 @@ append_extra_args_ascii(uint8_t **data, size_t *data_size)
 	if (!new_data)
 		return 0;
 
-	if (*data)
-		free(*data);
 	*data = (uint8_t *)new_data;
 	*data_size = usedchars;
 
@@ -790,7 +829,7 @@ append_extra_args_unicode(uint8_t **data, size_t *data_size)
 	int i;
 	unsigned long usedchars=0;
 
-	if (!data)
+	if (!data || *data)
 		return -1;
 
 	for (i = opts.optind; i < opts.argc; i++) {
@@ -816,8 +855,6 @@ append_extra_args_unicode(uint8_t **data, size_t *data_size)
 	if (!new_data)
 		return 0;
 
-	if (*data)
-		free(*data);
 	*data = (uint8_t *)new_data;
 	*data_size = usedchars * sizeof (*new_data);
 
@@ -832,7 +869,7 @@ append_extra_args_file(uint8_t **data, size_t *data_size)
 	ssize_t num_read=0;
 	unsigned long appended=0;
 	size_t maxchars = 0;
-	char *buffer; 
+	char *buffer;
 
 	if (!data) {
 		fprintf(stderr, "internal error\n");
@@ -906,6 +943,7 @@ append_extra_args(uint8_t **data, size_t *data_size)
 		free(new_data);
 		if (ret < 0)
 			return -1;
+		new_data = NULL;
 		new_data_size = 0;
 	}
 
@@ -913,15 +951,21 @@ append_extra_args(uint8_t **data, size_t *data_size)
 		ret = append_extra_args_unicode(&new_data, &new_data_size);
 	else
 		ret = append_extra_args_ascii(&new_data, &new_data_size);
-	if (ret < 0)
+	if (ret < 0) {
+		if (new_data) /* this can't happen, but covscan believes */
+			free(new_data);
 		return -1;
+	}
 	if (new_data_size) {
 		ret = add_new_data(data, data_size, new_data, new_data_size);
 		free(new_data);
+		new_data = NULL;
 		if (ret < 0)
 			return -1;
 		new_data_size = 0;
 	}
 
+	if (new_data) /* once again, this can't happen, but covscan believes */
+		free(new_data);
 	return 0;
 }
