@@ -221,6 +221,7 @@ read_lba(int fd, uint64_t lba, void *buffer, size_t bytes)
         void *iobuf;
         size_t iobuf_size;
         int rc;
+	off_t new_offset;
 
         iobuf_size = lcm(bytes, sector_size);
         rc = posix_memalign(&iobuf, sector_size, iobuf_size);
@@ -228,8 +229,11 @@ read_lba(int fd, uint64_t lba, void *buffer, size_t bytes)
                 return rc;
         memset(iobuf, 0, bytes);
 
-
-        lseek(fd, offset, SEEK_SET);
+        new_offset = lseek(fd, offset, SEEK_SET);
+	if (new_offset == (off_t)-1) {
+		free(iobuf);
+		return 0;
+	}
         bytesread = read(fd, iobuf, iobuf_size);
         memcpy(buffer, iobuf, bytes);
         free(iobuf);
@@ -608,6 +612,7 @@ gpt_disk_get_partition_info(int fd,
 {
 	gpt_header *gpt = NULL;
 	gpt_entry *ptes = NULL, *p;
+	int rc = 0;
 
 	if (!find_valid_gpt(fd, &gpt, &ptes))
 		return 1;
@@ -624,9 +629,14 @@ gpt_disk_get_partition_info(int fd,
 		       sizeof (p->unique_partition_guid));
 	} else {
 		fprintf (stderr,"partition %d is not valid\n", num);
-		return 1;
+		rc = 1;
 	}
-	return 0;
+	if (ptes)
+		free(ptes);
+	if (gpt)
+		free(gpt);
+
+	return rc;
 }
 
 /*
